// Copyright Oryx Mail Systems GmbH. All enquiries to info@oryx.com, please.

#include "frontmatter.h"

#include "configuration.h"
#include "pagecomponent.h"


class FrontMatterData
    : public Garbage
{
public:
    EString element;
};


/*! \class FrontMatter frontmatter.h
    Provides front matter needed to render the rest of a WebPage.

    This class inherits from EString and provides a collection of static
    functions that return a pointer to a new FrontMatter object, whose
    contents are an HTML string suitable for inclusion into a WebPage.

    PageComponent constructors call PageComponent::addFrontMatter() on
    the FrontMatter objects they need, and the WebPage includes their
    contents (wrapped appropriately, according to the element() name)
    in the <head> section while rendering itself.
*/

/*! Creates a new FrontMatter object for the specified \a element. */

FrontMatter::FrontMatter( const EString & element )
    : d( new FrontMatterData )
{
    d->element = element;
}


/*! Returns this object's element name, as specified to the
    constructor. */

EString FrontMatter::element() const
{
    return d->element;
}


/*! Returns a title element for \a s, which will be HTML quoted. */

FrontMatter * FrontMatter::title( const EString & s )
{
    FrontMatter * fm = new FrontMatter( "title" );
    fm->append( PageComponent::quoted( s ) );
    return fm;
}


/*! Returns a link to the stylesheet. */

FrontMatter * FrontMatter::styleSheet()
{
    FrontMatter * fm = new FrontMatter( "style" );

    fm->append( "@import url(\"" );
    fm->append( Configuration::text( Configuration::WebmailCSS ) );
    fm->append( "\");\n" );

    // The following classes are used to display different content
    // depending on whether JavaScript is active.
    //
    // - .jsonly is visible only if JS is active.
    // - .njsvisible is visible only if JS is inactive.
    // - .hidden is invisible.
    // - .njshidden is invisible too.
    //
    // During page load, the JavaScript code changes the js and njs
    // classes so that they act as described.

    fm->append( ".jsonly{display:none;}\n"
                ".njsvisible{}\n"
                ".hidden{display:none;}\n"
                ".njshidden{display:none;}\n" );

    return fm;
}


/*! Returns an object that contains \a s and belongs inside a <style>
    element. */

FrontMatter * FrontMatter::style( const EString &s )
{
    FrontMatter * fm = new FrontMatter( "style" );
    fm->append( s );
    fm->append( "\n" );
    return fm;
}


/*! Returns a link to the (compressed) jquery source. */

FrontMatter * FrontMatter::jQuery()
{
    FrontMatter * fm = new FrontMatter( "script" );
    fm->append( "<script type=\"text/javascript\" "
                "src=\"http://ajax.googleapis.com/"
                "ajax/libs/jquery/1.2.6/jquery.min.js\">"
                "</script>\n" );
    return fm;
}


/*! Returns a <script> element wrapping the given code \a s. */

FrontMatter * FrontMatter::script( const EString &s )
{
    FrontMatter * fm = new FrontMatter( "script" );
    fm->append( s );
    fm->append( "\n" );
    return fm;
}


/*! Returns a glob of JavaScript code. */

FrontMatter * FrontMatter::jsToggles()
{
    FrontMatter * fm = new FrontMatter( "script" );

    // Define a useJS function to change the stylesheet to make the js
    // and njs classes work if JavaScript is enabled.
    fm->append( "var toggledToJs=false;\n"
                "function useJS(){\n"
                "if(toggledToJs) return;\n"
                "var r=new Array;\n"
                "if(document.styleSheets[0].cssRules)"
                "r=document.styleSheets[0].cssRules;\n"
                "else if(document.styleSheets[0].rules)"
                "r=document.styleSheets[0].rules;\n"
                "else return;\n"
                "var i=0;\n"
                "if(r[1].style.display=='none')"
                "i=1;\n"
                "r[i].style.display='';\n"
                "r[i+1].style.display='none';\n"
                "toggledToJs=true\n"
                "}\n" );

    // Call useJS at once (for browsers where we can modify the
    // stylesheet before the import has been completed), and in
    // window.onload for other browsers.
    fm->append( "useJS(); window.onload = 'useJS();';\n" );

    // A function to show an element
    fm->append( "function reveal(e){\n"
                "document.getElementById(e).className='visible';\n"
                "}\n" );

    // A function to hide an element
    fm->append( "function hide(e){\n"
                "document.getElementById(e).className='hidden';\n"
                "}\n" );

    // A function to set a button's text. Does not work on any other
    // HTML elements.
    fm->append( "function setButtonText(i,t){\n"
                "var e = document.getElementById(i);\n"
                "if(e){\n"
                "e.childNodes[0].data=t;\n"
                "}\n"
                "}\n" );

    // A function to expand/collapse a message
    fm->append( "var hiddenIds=new Array;\n"
                "function expandCollapse(i,a,b,c){\n"
                "if(hiddenIds[i]){\n"
                "reveal(a);\n"
                "reveal(b);\n"
                "hide(c);\n"
                "hiddenIds[i]=false\n"
                "}else{\n"
                "hide(a);\n"
                "hide(b);\n"
                "reveal(c);\n"
                "hiddenIds[i]=true\n"
                "}\n"
                "}\n" );

    return fm;
}
